package cz.fidentis.analyst.visitors.mesh;

import cz.fidentis.analyst.mesh.MeshVisitor;
import cz.fidentis.analyst.mesh.core.MeshFacet;
import cz.fidentis.analyst.mesh.core.MeshPoint;
import cz.fidentis.analyst.mesh.core.MeshTriangle;
import cz.fidentis.analyst.mesh.core.TriangleFan;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import javax.vecmath.Vector3d;

/**
 * Abstract class for algorithms calculating curvatures of mesh vertices.
 * @see https://computergraphics.stackexchange.com/questions/1718/what-is-the-simplest-way-to-compute-principal-curvature-for-a-mesh-triangle
 * @see http://rodolphe-vaillant.fr/?e=20
 * <p>
 * <b>All curvature algorithms suppose that the triangle vertices are oriented clockwise!</b>
 * </p>
 * 
 * @author Natalia Bebjakova
 * @author Radek Oslejsek
 */
public class Curvature extends MeshVisitor {
    
    private final Map<MeshFacet, List<Double>> gaussian = new HashMap<>();
    private final Map<MeshFacet, List<Double>> mean = new HashMap<>();
    private final Map<MeshFacet, List<Double>> minPrincipal = new HashMap<>();
    private final Map<MeshFacet, List<Double>> maxPrincipal = new HashMap<>();
    
    /**
     * Returns Gaussian curvatures for all inspected mesh facets. The order corresponds to
     * the order of vertices, i.e., i-th value represents the curvature of i-th mesh vertex.
     * 
     * @return Gaussian curvatures of inspected mesh facets.
     */
    public Map<MeshFacet, List<Double>> getGaussianCurvatures() {
        return Collections.unmodifiableMap(gaussian);
    }
    
    /**
     * Returns mean curvatures for all inspected mesh facets. The order corresponds to
     * the order of vertices, i.e., i-th value represents the curvature of i-th mesh vertex.
     * 
     * @return Mean curvatures of inspected mesh facets.
     */
    public Map<MeshFacet, List<Double>> getMeanCurvatures() {
        return Collections.unmodifiableMap(mean);
    }
    
    /**
     * Returns minimum principal curvatures for all inspected mesh facets. The order corresponds to
     * the order of vertices, i.e., i-th value represents the curvature of i-th mesh vertex.
     * 
     * @return Minimum principal curvatures of inspected mesh facets.
     */
    public Map<MeshFacet, List<Double>> getMinPrincipalCurvatures() {
        return Collections.unmodifiableMap(minPrincipal);
    }
    
    /**
     * Returns maximum principal curvatures for all inspected mesh facets. The order corresponds to
     * the order of vertices, i.e., i-th value represents the curvature of i-th mesh vertex.
     * 
     * @return Maximum principal curvatures of inspected mesh facets.
     */
    public Map<MeshFacet, List<Double>> getMaxPrincipalCurvatures() {
        return Collections.unmodifiableMap(maxPrincipal);
    }
    
    @Override
    public void visitMeshFacet(final MeshFacet facet) {
        synchronized (this) {
            if (gaussian.containsKey(facet)) {
                return; // already visited facet
            }
            gaussian.put(facet, new ArrayList<>());
            mean.put(facet, new ArrayList<>());
            minPrincipal.put(facet, new ArrayList<>());
            maxPrincipal.put(facet, new ArrayList<>());
        }
        
        Cache cache = new Cache(facet);
        
        //final List<CurvTriangle> triangles = precomputeTriangles(facet);
        for (int vertA = 0; vertA < facet.getNumberOfVertices(); vertA++) {
            TriangleFan oneRing = facet.getOneRingNeighborhood(vertA);
            
            if (oneRing.isBoundary()) {
                this.gaussian.get(facet).add(0.0);
                this.mean.get(facet).add(0.0);
                this.minPrincipal.get(facet).add(0.0);
                this.maxPrincipal.get(facet).add(0.0);
                continue;
            }
        
            double sumAngles = 0.0;
            double sumArea = 0.0;
            Vector3d pointSum = new Vector3d();
            
            for (int i = 0; i < oneRing.getVertices().size()-1; i++) {
                int vertB = oneRing.getVertices().get(i);
                int vertC = oneRing.getVertices().get(i+1);
                int tri = oneRing.getTriangles().get(i);
                
                // Angles:
                sumAngles += cache.getAngle(vertA, vertB, vertC, tri);
                
                // Area:
                double alpha = cache.getAngle(vertA, vertB, vertC, tri);
                double beta = cache.getAngle(vertB, vertA, vertC, tri);
                double gamma = cache.getAngle(vertC, vertA, vertB, tri);
                
                Vector3d ab = new Vector3d(facet.getVertex(vertB).getPosition());
                ab.sub(facet.getVertex(vertA).getPosition());
        
                Vector3d ac = new Vector3d(facet.getVertex(vertC).getPosition());
                ac.sub(facet.getVertex(vertA).getPosition());
        
                double lenSqrtAB = ab.lengthSquared();
                double lenSqrtAC = ac.lengthSquared();
                
                double piHalf = Math.PI / 2.0; // 90 degrees
                if (alpha >= piHalf || beta >= piHalf || gamma >= piHalf) { // check for obtuse angle
                    double area = 0.5 * lenSqrtAB * lenSqrtAC * Math.sin(alpha);
                    sumArea += (alpha > piHalf) ? area / 2.0 : area / 4.0;
                } else {
                    double cotGamma = 1.0 / Math.tan(gamma);
                    double cotBeta = 1.0 / Math.tan(beta);
                    sumArea += (lenSqrtAB * cotGamma + lenSqrtAC * cotBeta) / 8.0;
                }
                
                // Laplace:
                int triNext = oneRing.getTriangles().get((i + 1) % oneRing.getTriangles().size());
                int vertD;
                if (i == (oneRing.getVertices().size() - 2)) { // last triangle
                    vertD = oneRing.getVertices().get(1);
                } else {
                    vertD = oneRing.getVertices().get(i + 2);
                }
                ac.scale(cache.getCtan(vertA, vertB, vertC, tri) + cache.getCtan(vertA, vertC, vertD, triNext));
                pointSum.add(ac);
            }
        
            double gaussVal = (2.0 * Math.PI - sumAngles) / sumArea;
            double meanVal = 0.25 * sumArea * pointSum.length();
            double delta = Math.max(0, Math.pow(meanVal, 2) - gaussVal);
            
            this.gaussian.get(facet).add(gaussVal);
            this.mean.get(facet).add(meanVal);
            this.minPrincipal.get(facet).add(meanVal - Math.sqrt(delta));
            this.maxPrincipal.get(facet).add(meanVal + Math.sqrt(delta));
        }
    }
    
    /*
    protected List<CurvTriangle> precomputeTriangles(MeshFacet facet) {
        List<CurvTriangle> ret = new ArrayList<>(facet.getNumTriangles());
        for (MeshTriangle tri: facet) {
            ret.add(new CurvTriangle(facet, tri));
        }
        return ret;
    }
    */
    
    /**
     * Helper class that caches triangle characteristics used multiples times during the curvature computation.
     * <ul>
     * <li>A = central point of the 1-ring neighborhood</li>
     * <li>B = point "on the left" (previous point in the clockwise orientation of the triangle</li>
     * <li>C = point "on the right" (next point in the clockwise orientation of the triangle</li>
     * </ul>
     * 
     * @author Radek Oslejsek
     */
    protected class CurvTriangle {
        
    }

    /**
     * Cache key.
     * @author Radek Oslejsek
     */
    private class CacheKey {

        private final int vertIndex;
        private final int triIndex;
        
        /**
         * Constructor.
         * @param v vertex
         * @param i triangle
         */
        CacheKey(int v, int i) {
            this.vertIndex = v;
            this.triIndex = i;
        }

        @Override
        public int hashCode() {
            int hash = 7;
            hash = 61 * hash + this.vertIndex;
            hash = 61 * hash + this.triIndex;
            return hash;
        }

        @Override
        public boolean equals(Object obj) {
            if (this == obj) {
                return true;
            }
            if (obj == null) {
                return false;
            }
            if (getClass() != obj.getClass()) {
                return false;
            }
            final CacheKey other = (CacheKey) obj;
            if (this.vertIndex != other.vertIndex) {
                return false;
            }
            if (this.triIndex != other.triIndex) {
                return false;
            }
            return true;
        }
    }
    
    /**
     * Cache.
     * @author Radek Oslejsek
     */
    private class Cache {
        
        private final MeshFacet facet;
        private final Map<CacheKey, Double> cache = new HashMap<>();

        Cache(MeshFacet facet) {
            this.facet = facet;
        }
        
        public double getAngle(int vertA, int vertB, int vertC, int tri) {
            CacheKey key = new CacheKey(vertA, tri);
            if (cache.containsKey(key)) {
                return cache.get(key);
            }
            
            //System.out.println("XXX"+vertA+"|"+vertB+"|"+vertC);
            //System.out.println("YYY"+facet.getVertex(vertA).getPosition()+"|"+facet.getVertex(vertB).getPosition()+"|"+facet.getVertex(vertC).getPosition());
            
            Vector3d ab = new Vector3d(facet.getVertex(vertB).getPosition());
            ab.sub(facet.getVertex(vertA).getPosition());
            ab.normalize();
            
            Vector3d ac = new Vector3d(facet.getVertex(vertC).getPosition());
            ac.sub(facet.getVertex(vertA).getPosition());
            ac.normalize();
            
            Vector3d bc = new Vector3d(facet.getVertex(vertC).getPosition());
            bc.sub(facet.getVertex(vertB).getPosition());
            bc.normalize();
            
            //System.out.println("AAA"+ac+"|"+ab+"|"+bc);
            
            double cosAlpha = ab.dot(ac);
            
            ab.scale(-1.0);
            double cosBeta = ab.dot(bc);
            
            bc.scale(-1.0);
            ac.scale(-1.0);
            double cosGamma = bc.dot(ac);
            
            //System.out.println("BBB"+cosAlpha+"|"+cosBeta+"|"+cosGamma);
            
            double ret = Math.acos(cosAlpha);
            
            cache.put(new CacheKey(vertA, tri), ret);
            cache.put(new CacheKey(vertB, tri), Math.acos(cosBeta));
            cache.put(new CacheKey(vertC, tri), Math.acos(cosGamma));
            
            return ret;
        }
        
        public double getCtan(int vertA, int vertB, int vertC, int tri) {
            return 1.0 / Math.tan(getAngle(vertA, vertB, vertC, tri));
        }
    }
    

}
